--Base functions

function IsAnyActorAtPosition(Position)
	local Result=nil;
	local actorID = MovableMan:GetRootMOID(SceneMan:GetMOIDPixel(Position.X,Position.Y));
	for actor in MovableMan.Actors do
		if actor.ID == actorID then
			Result = actor;
			break;
		end
	end
	return Result;
end

function ConvertAngleToNormalizedVector(Angle)
	return Vector(math.cos(Angle),-math.sin(Angle));
end

function ReturnRandomPositiveNegative(Number)
	if Number == nil then
		Number = 1;
	end
	Number = Number-math.random()*Number*2;
	return Number;
end

function RandomizeParticlePerpendicularVelocity(Particle, MaximimVelocity, VectorToDestination)
	Particle.Vel= VectorToDestination.Normalized.Perpendicular*ReturnRandomPositiveNegative(MaximimVelocity);
end

function RandomizeForSmokeTrace(Particle, MaximimVelocity, VectorToDestination)
	Particle.Vel= VectorToDestination.Normalized.Perpendicular*ReturnRandomPositiveNegative(MaximimVelocity);
	Particle.Age = math.random(350);
end

function RandomizeForWarper(Particle, Aging, VectorToDestination, ParticleNumber)
	--Particle.Vel= VectorToDestination.Normalized.Perpendicular*ReturnRandomPositiveNegative(MaximimVelocity);
	Particle.Age = ParticleNumber*Aging;
end

function RandomizeForWarperRings(Particle, MaximimVelocity, VectorToDestination, ParticleNumber)
	Particle.Vel= VectorToDestination.Normalized.Perpendicular*MaximimVelocity+VectorToDestination.Normalized.Perpendicular*ReturnRandomPositiveNegative(0.35);
end

function DrawLine(StartPosition, EndPosition, DrawLengthMultiplier, DrawingQuality, MOPixelToDrawWith)	
	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true) * DrawLengthMultiplier;
	local ParticleCount = VectorToDestination.Magnitude * DrawingQuality;
	local DistanceMultiplier = (1 / DrawingQuality);
	for ParticleNumber = 0, ParticleCount, 1 do
		local Particle = CreateMOPixel(MOPixelToDrawWith);
		Particle.Pos = StartPosition+ VectorToDestination.Normalized * DistanceMultiplier * ParticleNumber;
		MovableMan:AddParticle(Particle);
	end
end

function DrawChaoticLine(StartPosition, EndPosition, MaximumDeviationFromTheAxis, DrawLengthMultiplier, DrawingQuality, MOPixelToDrawWith)	
	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true) * DrawLengthMultiplier;
	local ParticleCount = VectorToDestination.Magnitude * DrawingQuality;
	local DistanceMultiplier = (1 / DrawingQuality);
	for ParticleNumber = 0, ParticleCount, 1 do
		local Particle = CreateMOPixel(MOPixelToDrawWith);
		local Offset = VectorToDestination.Normalized.Perpendicular * ReturnRandomPositiveNegative(MaximumDeviationFromTheAxis);
		Particle.Pos = StartPosition + VectorToDestination.Normalized * ParticleNumber * DistanceMultiplier + Offset;
		MovableMan:AddParticle(Particle);
	end
end

function DrawChaoticLineWithPostProcessing(StartPosition, EndPosition, MaximumDeviationFromTheAxis,DrawLengthMultiplier, DrawingQuality, MOPixelToDrawWith, Process, ProcessParameter)	
	if	Process == nil then 
		DrawChaoticLine(StartPosition, EndPosition, MaximumDeviationFromTheAxis, DrawLengthMultiplier, DrawingQuality, MOPixelToDrawWith);
	else
		local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true) * DrawLengthMultiplier;
		local ParticleCount = VectorToDestination.Magnitude * DrawingQuality;
		local DistanceMultiplier = (1 / DrawingQuality);
		for ParticleNumber = 0, ParticleCount, 1 do
			local Particle = CreateMOPixel(MOPixelToDrawWith);
			local Offset = VectorToDestination.Normalized.Perpendicular * ReturnRandomPositiveNegative(MaximumDeviationFromTheAxis);
			Particle.Pos = StartPosition + VectorToDestination.Normalized * ParticleNumber * DistanceMultiplier + Offset;
			MovableMan:AddParticle(Particle);
			pcall(Process, Particle, ProcessParameter, VectorToDestination,ParticleNumber, ParticleCount);
		end
	end
end

--function DrawMPChaoticLineWithPostProcessing(StartPosition, EndPosition, MaximumDeviationFromTheAxis,DrawLengthMultiplier, DrawingQuality, MOPixelToDrawWithList, Process, ProcessParameter)	
--	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true) * DrawLengthMultiplier;
--	local ParticleCount = VectorToDestination.Magnitude * DrawingQuality;
--	local DistanceMultiplier = (1 / DrawingQuality);
--	local ParticlesInStage = ParticleCount / #MOPixelToDrawWithList;
--	for ParticleNumber = 0, ParticleCount, 1 do
--		local Stage = math.ceil(ParticleNumber / ParticlesInStage)+1;
--		local MOPixelToDrawWith = MOPixelToDrawWithList[Stage];
--		local Particle = CreateMOPixel(MOPixelToDrawWith);
--		local Offset = VectorToDestination.Normalized.Perpendicular * ReturnRandomPositiveNegative(MaximumDeviationFromTheAxis);
--		Particle.Pos = StartPosition + VectorToDestination.Normalized * ParticleNumber * DistanceMultiplier + Offset;
--		MovableMan:AddParticle(Particle);
--		pcall(Process, Particle, ProcessParameter, VectorToDestination,ParticleNumber, ParticleCount);
--	end
--end

function BasicFractureVector(StartPosition,EndPosition,FragmentCount,Deviation)
	local Path = {};
	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true);
	local Step = VectorToDestination / FragmentCount;
	local Position = StartPosition;
	
	Path[0] = StartPosition;
	Path[FragmentCount] = EndPosition;
	for FragmentNumber = 1, FragmentCount-1, 1 do
		local Offset = VectorToDestination.Normalized.Perpendicular * ReturnRandomPositiveNegative(Deviation);
		Path[FragmentNumber] = Position + Step + Offset;
		Position = Position + Step;
	end
	return Path;
end

function ScanVectorForMO(StartPosition, VectorToDestination, MOIDToIgnore)
	local ResultMOID = 255;
	local ResultPosition = StartPosition + VectorToDestination;
	local CheckMO = SceneMan:CastMORay(StartPosition, VectorToDestination, MOIDToIgnore, -1, false, 1);
	if CheckMO ~= 255 then
		local MO = MovableMan:GetMOFromID(CheckMO);
		if MO.GetsHitByMOs then
			ResultMOID = CheckMO;
			ResultPosition = StartPosition + (VectorToDestination/ VectorToDestination.Magnitude * SceneMan:ShortestDistance(StartPosition,MO.Pos,true).Magnitude);
		end	
	end
	return ResultPosition, ResultMOID;
end

function ScanVectorForMOTerrainAware(StartPosition, EndPosition, MOIDToIgnore)
	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true);
	local ResultPosition = EndPosition;
	local ResultMOID = 255;
	local ResultTerrainContact = false;
	CheckTerrain = SceneMan:CastNotMaterialRay(StartPosition, VectorToDestination, 0, 1,false);
	if CheckTerrain < 0 then
		ResultPosition, ResultMOID = ScanVectorForMO(StartPosition, VectorToDestination, MOIDToIgnore);
	else
		VectorToDestination = VectorToDestination.Normalized * CheckTerrain;
		ResultPosition, ResultMOID = ScanVectorForMO(StartPosition, VectorToDestination, MOIDToIgnore);
		if ResultMOID == 255 then
			ResultPosition = StartPosition + VectorToDestination;
			ResultTerrainContact = true;
		end
	end
	return ResultPosition, ResultMOID, ResultTerrainContact;
end

function ScanSectorForMOTerrainAware(StartPosition, ScanAxisVector, Radius, DegreesToScan, ScanStepDegrees, MOIDToIgnore)
	local StepCount = math.ceil(DegreesToScan / ScanStepDegrees);
	local ScanVector = Vector(ScanAxisVector.X,ScanAxisVector.Y):DegRotate(-(DegreesToScan/2));
	local ResultPosition = nil
	local ResultMOID = 255
	for StepNumber = 0, math.ceil(DegreesToScan / ScanStepDegrees), 1 do
		ScanVector = ScanVector:DegRotate(ScanStepDegrees):SetMagnitude(Radius);
		--DrawLine(StartPosition, StartPosition+ScanVector, 1, 0.1, "LBoltBlueLightTiny")
		ResultPosition, ResultMOID = ScanVectorForMOTerrainAware(StartPosition,StartPosition + ScanVector,MOIDToIgnore);
		if ResultMOID ~= 255 then
			break;
		end
	end
	return ResultPosition, ResultMOID;
end

function ScanSectorForActorTerrainAware(StartPosition, ScanAxisVector, Radius, DegreesToScan, ScanStepDegrees, MOIDToIgnore)
	local StepCount = math.ceil(DegreesToScan / ScanStepDegrees);
	local ScanVector = Vector(ScanAxisVector.X,ScanAxisVector.Y):DegRotate(-(DegreesToScan/2));
	local ResultPosition = nil
	local ResultMOID = 255
	for StepNumber = 0, math.ceil(DegreesToScan / ScanStepDegrees), 1 do
		ScanVector = ScanVector:DegRotate(ScanStepDegrees):SetMagnitude(Radius);
		--DrawLine(StartPosition, StartPosition+ScanVector, 1, 0.1, "LBoltBlueLightTiny")
		ResultPosition, ResultMOID = ScanVectorForMOTerrainAware(StartPosition,StartPosition + ScanVector,MOIDToIgnore);
		if ResultMOID ~= 255 and  MovableMan:GetMOFromID(ResultMOID).Class == "AHuman" then
			break;
		end
	end
	return ResultPosition, ResultMOID;
end

function FractureVectorConsideringTerrainAndMOs(StartPosition,EndPosition,FragmentCount,Deviation,MOIDToIgnore)
	local Path = {};
	local VectorToDestination = SceneMan:ShortestDistance(StartPosition,EndPosition,true);
	local Step = VectorToDestination / FragmentCount;
	local Position = StartPosition;
	local ResultPosition;
	local ResultMOID = 255;
	local ResultTerrain = false;
	
	Path[0] = StartPosition;
	--Path[FragmentCount] = EndPosition;
	for FragmentNumber = 1, FragmentCount, 1 do
		if FragmentNumber ~= FragmentCount then
			local Offset = VectorToDestination.Normalized.Perpendicular * ReturnRandomPositiveNegative(Deviation);
			ResultPosition, ResultMOID, ResultTerrain = ScanVectorForMOTerrainAware(Position, Position + Step + Offset, MOIDToIgnore);
		else
			ResultPosition, ResultMOID, ResultTerrain = ScanVectorForMOTerrainAware(Position, Position + Step, MOIDToIgnore);
		end
		Path[FragmentNumber] = ResultPosition;
		if ResultMOID == 255 and  ResultTerrain == false then
			Position = Position + Step;
		else
			break
		end;
	end
	return Path, ResultMOID, ResultTerrain;
end

function CastParticleStream(StartPosition,DirectionVector,ArcLength,MinimumDeviationFromTheAxis,MaximumDeviationFromTheAxis,MinimumNumberOfArcFractures,MaximumNumberOfArcFractures,MOPixelToDrawWith,DrawJitter,MinimumDrawLengthMultiplier,MaximumDrawLengthMultiplier,DrawingQuality,MOIDToIgnore,FunctionToCallOnHitMO,p1,FunctionToCallOnHitTerrain,p2,FunctionToCallOnHitNothing,p3,PassAllParameters)
	local VectorToDestination = DirectionVector/DirectionVector.Magnitude*ArcLength;
	local Path;
	local HitMO = 255;
	local HitTerrain = false;
	local NumberOfArcFractures = math.min(MaximumNumberOfArcFractures,MinimumNumberOfArcFractures)+math.ceil(math.random()*math.abs(MaximumNumberOfArcFractures-MinimumNumberOfArcFractures));
	local DeviationFromTheAxis = math.min(MaximumDeviationFromTheAxis,MinimumDeviationFromTheAxis)+math.ceil(math.random()*math.abs(MaximumDeviationFromTheAxis-MinimumDeviationFromTheAxis));

	Path,HitMO,HitTerrain = FractureVectorConsideringTerrainAndMOs(StartPosition,StartPosition+VectorToDestination,NumberOfArcFractures,DeviationFromTheAxis,MOIDToIgnore);		
	local NumberOfSteps = #Path - 1;
	for Step=0,NumberOfSteps,1 do	
		local DrawLengthMultiplier = math.min(MaximumDrawLengthMultiplier,MinimumDrawLengthMultiplier)+(math.random()*math.abs(MaximumDrawLengthMultiplier-MinimumDrawLengthMultiplier));
		if Step == NumberOfSteps then
			DrawLengthMultiplier = 1;
		end
		if DrawJitter == 0 then
			DrawLine(Path[Step],Path[Step + 1],DrawLengthMultiplier,DrawingQuality,MOPixelToDrawWith);
		else
			DrawChaoticLine(Path[Step],Path[Step + 1],DrawJitter,DrawLengthMultiplier,DrawingQuality,MOPixelToDrawWith);
		end
	end
	
	if HitMO ~= 255 then
		if PassAllParameters == true then 
			pcall(FunctionToCallOnHitMO,Path,HitMO,HitTerrain,p1,StartPosition,DirectionVector,ArcLength,MinimumDeviationFromTheAxis,MaximumDeviationFromTheAxis,MinimumNumberOfArcFractures,MaximumNumberOfArcFractures,MOPixelToDrawWith,DrawJitter,MinimumDrawLengthMultiplier,MaximumDrawLengthMultiplier,DrawingQuality,MOIDToIgnore,FunctionToCallOnHitMO,p1,FunctionToCallOnHitTerrain,p2,FunctionToCallOnHitNothing,p3);
		else
			pcall(FunctionToCallOnHitMO,Path,HitMO,HitTerrain,p1);
		end
	else
		if (HitTerrain ~= false) then
			if PassAllParameters == true then 
				pcall(FunctionToCallOnHitTerrain,StartPosition,DirectionVector,ArcLength,MinimumDeviationFromTheAxis,MaximumDeviationFromTheAxis,MinimumNumberOfArcFractures,MaximumNumberOfArcFractures,MOPixelToDrawWith,DrawJitter,MinimumDrawLengthMultiplier,MaximumDrawLengthMultiplier,DrawingQuality,MOIDToIgnore,FunctionToCallOnHitMO,p1,FunctionToCallOnHitTerrain,p2,FunctionToCallOnHitNothing,p3,Path,HitMO,HitTerrain,p2);
			else
				pcall(FunctionToCallOnHitTerrain,Path,HitMO,HitTerrain,p2);
			end
		else
			if PassAllParameters == true then 
				pcall(FunctionToCallOnHitNothing,StartPosition,DirectionVector,ArcLength,MinimumDeviationFromTheAxis,MaximumDeviationFromTheAxis,MinimumNumberOfArcFractures,MaximumNumberOfArcFractures,MOPixelToDrawWith,DrawJitter,MinimumDrawLengthMultiplier,MaximumDrawLengthMultiplier,DrawingQuality,MOIDToIgnore,FunctionToCallOnHitMO,p1,FunctionToCallOnHitTerrain,p2,FunctionToCallOnHitNothing,p3,Path,HitMO,HitTerrain,p3);
			else
				pcall(FunctionToCallOnHitNothing,Path,HitMO,HitTerrain,p3);
			end
		end
	end

end

--On hit effects

function DoNothing()
end

function Explosion(Pos,Vel)
	local expl = CreateMOSRotating("LDischarge");
	expl.Pos = Pos;
	if Vel ~= nil then
		expl.Vel = Vel;
	end
	MovableMan:AddParticle(expl);
end

function GibMOSRotating(Pos,Vel,MOSRotatingName)
	local expl = CreateMOSRotating(MOSRotatingName);
	expl.Pos = Pos;
	if Vel ~= nil then
		expl.Vel = Vel;
	end
	MovableMan:AddParticle(expl);
	expl:GibThis();
end

function LastPositionExplosion(Path)
	Explosion(Path[#Path]);
end

function ExplosionOnRootMO(Path,HitMO)
	if HitMO ~= 255 and MovableMan:GetMOFromID(HitMO).GetsHitByMOs then
		local RootID = MovableMan:GetRootMOID(HitMO);
		local RootMO = MovableMan:GetMOFromID(RootID)
		DrawChaoticLine(Path[#Path],RootMO.Pos,0.5,1,1,"LBoltBlueLightTiny");
		Explosion(RootMO.Pos,RootMO.Vel);
	end
end

function ExplosionOnRootMO10(Path,HitMO)
	if HitMO ~= 255 and MovableMan:GetMOFromID(HitMO).GetsHitByMOs then
		local RootID = MovableMan:GetRootMOID(HitMO);
		local RootMO = MovableMan:GetMOFromID(RootID)
		for i=0,10,1 do
			Explosion(RootMO.Pos,RootMO.Vel);
		end
	end
end

function LStrikeJunction(Path)
--CHECK WITHIN BOUNDS???!!!
	for i=0,math.random(2),1 do
		LStrike(Path[#Path],Vector(ReturnRandomPositiveNegative(1),1),300,255);
	end
end

-- function Shock
function ChainLightning(Path,HitMO,HitTerrain,Parameter)
	ExplosionOnRootMO(Path,HitMO);
	local Particle = CreateMOPixel("ChainLightningSource");
	Particle.Mass = Parameter.X;
	Particle.Sharpness = Parameter.Y;
	Particle.Pos = Path[#Path];
	MovableMan:AddParticle(Particle);
end

--Presets for CastParticleStream

function CastD9Lightning(Pos,Aim,Length,MOIDToIgnore)
	CastParticleStream(Pos, Aim,Length,3,10,17,34,"LBoltBlueLightTiny",0.5,1,1.5,1,MOIDToIgnore,ExplosionOnRootMO,nil,LastPositionExplosion,nil,DoNothing,nil);
end

function CastD9LightningWithAutoTargeting(Pos,Aim,Length,MOIDToIgnore,TargetingConeDegrees)
	local ResultPosition; 
	local ResultMOID = 255;
	ResultPosition, ResultMOID = ScanSectorForMOTerrainAware(Pos, Aim, Length, TargetingConeDegrees, 3, MOIDToIgnore);
	--print(ResultMOID);
	if ResultMOID == 255 then
		CastParticleStream(Pos, Aim,Length,3,10,17,34,"LBoltBlueLightTiny",0.5,1,1.5,1,MOIDToIgnore,ExplosionOnRootMO,nil,LastPositionExplosion,nil,DoNothing,nil);
	else
		local NewVector = SceneMan:ShortestDistance(Pos,ResultPosition,true);
		CastParticleStream(Pos, NewVector, NewVector.Magnitude,3,7,15,30,"LBoltBlueLightTiny",0.5,1,1.5,1,MOIDToIgnore,ExplosionOnRootMO,nil,LastPositionExplosion,nil,DoNothing,nil);
	end
end

function CastDecorativeLightning(Pos,Aim,Length,MOIDToIgnore)
	CastParticleStream(Pos, Aim,Length,3,10,17,34,"LBoltBlueLightTiny",0.5,1,1.5,1,MOIDToIgnore,DoNothing,nil,DoNothing,nil,DoNothing,nil);
end

function CastBeam(Pos,Aim,Length,MOIDToIgnore)
	CastParticleStream(Pos, Aim,Length,0,0,1,1,"YellowBeam",0,1,1,0.8,MOIDToIgnore,LastPositionExplosion,nil,LastPositionExplosion,nil,DoNothing,nil);
end

function CastElectroBeam(Pos,Aim,Length,MOIDToIgnore)
	CastParticleStream(Pos, Aim,Length,0,0,1,1,"LBoltBlueLightTiny",1.6,1,1,1,MOIDToIgnore,ExplosionOnRootMO,nil,DoNothing,nil,DoNothing,nil);
end

function LStrike(Pos,Aim,Length,MOIDToIgnore)
	CastParticleStream(Pos, Aim,Length,5,13,5,15,"White8",0.6,1,1.6,0.2,MOIDToIgnore,ExplosionOnRootMO10,nil,DoNothing,nil,LStrikeJunction,nil);
end

function CastChainLightning(Pos,Aim,Length,MOIDToIgnore,Generations)
	if Generations >=1 then
		CastParticleStream(Pos, Aim,Length,3,10,17,34,"LBoltBlueLightTiny",0.5,1,1.5,1,MOIDToIgnore,ChainLightning,Vector(Generations-1,MOIDToIgnore),DoNothing,nil,DoNothing,nil);
	end
end



